@@ -1,10 +1,11 @@
 use std::{
-    collections::{binary_heap::PeekMut, hash_map, BinaryHeap, HashMap, VecDeque},
+    collections::{binary_heap::PeekMut, hash_map, BinaryHeap, VecDeque},
     convert::TryFrom,
     mem,
 };
 
 use bytes::BufMut;
+use fxhash::FxHashMap;
 use tracing::{debug, trace};
 
 use super::{
@@ -22,8 +23,8 @@ use crate::{
 pub struct StreamsState {
     pub(super) side: Side,
     // Set of streams that are currently open, or could be immediately opened by the peer
-    pub(super) send: HashMap<StreamId, Send>,
-    pub(super) recv: HashMap<StreamId, Recv>,
+    pub(super) send: FxHashMap<StreamId, Send>,
+    pub(super) recv: FxHashMap<StreamId, Recv>,
     pub(super) next: [u64; 2],
     // Locally initiated
     pub(super) max: [u64; 2],
@@ -50,7 +51,7 @@ pub struct StreamsState {
     /// Streams are only added to this list when a write fails.
     pub(super) connection_blocked: Vec<StreamId>,
     /// Connection-level flow control budget dictated by the peer
-    max_data: u64,
+    pub(super) max_data: u64,
     /// The initial receive window
     receive_window: u64,
     /// Limit on incoming data, which is transmitted through `MAX_DATA` frames
@@ -88,8 +89,8 @@ impl StreamsState {
     ) -> Self {
         let mut this = Self {
             side,
-            send: HashMap::default(),
-            recv: HashMap::default(),
+            send: FxHashMap::default(),
+            recv: FxHashMap::default(),
             next: [0, 0],
             max: [0, 0],
             max_remote: [max_remote_bi.into(), max_remote_uni.into()],
@@ -173,7 +174,6 @@ impl StreamsState {
         frame: frame::Stream,
         payload_len: usize,
     ) -> Result<ShouldTransmit, TransportError> {
-        trace!(id = %frame.id, offset = frame.offset, len = frame.data.len(), fin = frame.fin, "got stream");
         let stream = frame.id;
         self.validate_receive_id(stream).map_err(|e| {
             debug!("received illegal STREAM frame");
@@ -274,10 +274,12 @@ impl StreamsState {
             Some(ss) => ss,
             None => return,
         };
-        self.events
-            .push_back(StreamEvent::Stopped { id, error_code });
-        stream.stop(error_code);
-        self.on_stream_frame(false, id);
+
+        if stream.try_stop(error_code) {
+            self.events
+                .push_back(StreamEvent::Stopped { id, error_code });
+            self.on_stream_frame(false, id);
+        }
     }
 
     pub fn reset_acked(&mut self, id: StreamId) {
@@ -292,8 +294,23 @@ impl StreamsState {
         }
     }
 
-    pub fn can_send(&self) -> bool {
-        !self.pending.is_empty()
+    /// Whether any stream data is queued, regardless of control frames
+    pub fn can_send_stream_data(&self) -> bool {
+        // Reset streams may linger in the pending stream list, but will never produce stream frames
+        self.pending.iter().any(|level| {
+            level
+                .queue
+                .borrow()
+                .iter()
+                .any(|id| self.send.get(id).map_or(false, |s| !s.is_reset()))
+        })
+    }
+
+    /// Whether MAX_STREAM_DATA frames could be sent for stream `id`
+    pub fn can_send_flow_control(&self, id: StreamId) -> bool {
+        self.recv
+            .get(&id)
+            .map_or(false, |s| s.receiving_unknown_size())
     }
 
     pub fn write_control_frames(
@@ -433,6 +450,7 @@ impl StreamsState {
                 break;
             }
 
+            let num_levels = self.pending.len();
             let mut level = match self.pending.peek_mut() {
                 Some(x) => x,
                 None => break,
@@ -445,8 +463,11 @@ impl StreamsState {
             let id = match level.queue.get_mut().pop_front() {
                 Some(x) => x,
                 None => {
-                    PeekMut::pop(level);
-                    continue;
+                    debug_assert!(
+                        num_levels == 1,
+                        "An empty queue is only allowed for a single level"
+                    );
+                    break;
                 }
             };
             let stream = match self.send.get_mut(&id) {
@@ -471,13 +492,26 @@ impl StreamsState {
             if fin {
                 stream.fin_pending = false;
             }
+
             if stream.is_pending() {
                 if level.priority == stream.priority {
+                    // Enqueue for the same level
                     level.queue.get_mut().push_back(id);
                 } else {
-                    drop(level);
+                    // Enqueue for a different level. If the current level is empty, drop it
+                    if level.queue.borrow().is_empty() && num_levels != 1 {
+                        // We keep the last level around even in empty form so that
+                        // the next insert doesn't have to reallocate the queue
+                        PeekMut::pop(level);
+                    } else {
+                        drop(level);
+                    }
                     push_pending(&mut self.pending, id, stream.priority);
                 }
+            } else if level.queue.borrow().is_empty() && num_levels != 1 {
+                // We keep the last level around even in empty form so that
+                // the next insert doesn't have to reallocate the queue
+                PeekMut::pop(level);
             }
 
             let meta = frame::StreamMeta { id, offsets, fin };
@@ -1070,12 +1104,22 @@ mod tests {
             conn_state: &state,
         };
 
-        let reason = 0u32.into();
-        stream.state.received_stop_sending(id, reason);
-        assert_eq!(stream.write(&[]), Err(WriteError::Stopped(reason)));
+        let error_code = 0u32.into();
+        stream.state.received_stop_sending(id, error_code);
+        assert!(stream
+            .state
+            .events
+            .contains(&StreamEvent::Stopped { id, error_code }));
+        stream.state.events.clear();
+
+        assert_eq!(stream.write(&[]), Err(WriteError::Stopped(error_code)));
 
         stream.reset(0u32.into()).unwrap();
         assert_eq!(stream.write(&[]), Err(WriteError::UnknownStream));
+
+        // A duplicate frame is a no-op
+        stream.state.received_stop_sending(id, error_code);
+        assert!(stream.state.events.is_empty());
     }
 
     #[test]
@@ -1145,6 +1189,76 @@ mod tests {
         assert_eq!(meta[0].id, id_high);
         assert_eq!(meta[1].id, id_mid);
         assert_eq!(meta[2].id, id_low);
+
+        assert!(!server.can_send_stream_data());
+        assert_eq!(server.pending.len(), 1);
+    }
+
+    #[test]
+    fn requeue_stream_priority() {
+        let mut server = make(Side::Server);
+        server.set_params(&TransportParameters {
+            initial_max_streams_bidi: 3u32.into(),
+            initial_max_data: 1000u32.into(),
+            initial_max_stream_data_bidi_remote: 1000u32.into(),
+            ..Default::default()
+        });
+
+        let (mut pending, state) = (Retransmits::default(), ConnState::Established);
+        let mut streams = Streams {
+            state: &mut server,
+            conn_state: &state,
+        };
+
+        let id_high = streams.open(Dir::Bi).unwrap();
+        let id_mid = streams.open(Dir::Bi).unwrap();
+
+        let mut mid = SendStream {
+            id: id_mid,
+            state: &mut server,
+            pending: &mut pending,
+            conn_state: &state,
+        };
+        assert_eq!(mid.write(b"mid").unwrap(), 3);
+        assert_eq!(server.pending.len(), 1);
+
+        let mut high = SendStream {
+            id: id_high,
+            state: &mut server,
+            pending: &mut pending,
+            conn_state: &state,
+        };
+        high.set_priority(1).unwrap();
+        assert_eq!(high.write(&[0; 200]).unwrap(), 200);
+        assert_eq!(server.pending.len(), 2);
+
+        // Requeue the high priority stream to lowest priority. The initial send
+        // still uses high priority since it's queued that way. After that it will
+        // switch to low priority
+        let mut high = SendStream {
+            id: id_high,
+            state: &mut server,
+            pending: &mut pending,
+            conn_state: &state,
+        };
+        high.set_priority(-1).unwrap();
+
+        let mut buf = Vec::with_capacity(1000);
+        let meta = server.write_stream_frames(&mut buf, 40);
+        assert_eq!(meta.len(), 1);
+        assert_eq!(meta[0].id, id_high);
+
+        // After requeuing we should end up with 2 priorities - not 3
+        assert_eq!(server.pending.len(), 2);
+
+        // Send the remaining data. The initial mid priority one should go first now
+        let meta = server.write_stream_frames(&mut buf, 1000);
+        assert_eq!(meta.len(), 2);
+        assert_eq!(meta[0].id, id_mid);
+        assert_eq!(meta[1].id, id_high);
+
+        assert!(!server.can_send_stream_data());
+        assert_eq!(server.pending.len(), 1);
     }
 
     #[test]
@@ -1172,4 +1286,34 @@ mod tests {
         stream.stop(0u32.into()).unwrap();
         assert!(client.recv.get_mut(&id).is_none(), "stream is freed");
     }
+
+    // Verify that a stream that's been reset doesn't cause the appearance of pending data
+    #[test]
+    fn reset_stream_cannot_send() {
+        let mut server = make(Side::Server);
+        server.set_params(&TransportParameters {
+            initial_max_streams_uni: 1u32.into(),
+            initial_max_data: 42u32.into(),
+            initial_max_stream_data_uni: 42u32.into(),
+            ..Default::default()
+        });
+        let (mut pending, state) = (Retransmits::default(), ConnState::Established);
+        let mut streams = Streams {
+            state: &mut server,
+            conn_state: &state,
+        };
+
+        let id = streams.open(Dir::Uni).unwrap();
+        let mut stream = SendStream {
+            id,
+            state: &mut server,
+            pending: &mut pending,
+            conn_state: &state,
+        };
+        stream.write(b"hello").unwrap();
+        stream.reset(0u32.into()).unwrap();
+
+        assert_eq!(pending.reset_stream, &[(id, 0u32.into())]);
+        assert!(!server.can_send_stream_data());
+    }
 }
